import collections
import contextlib
import json
import threading
import uuid

from django.conf import settings
from django.db import connection
from django.utils.module_loading import import_string


_tracker = threading.local()


Context = collections.namedtuple("Context", ["id", "metadata"])


def _is_concurrent_statement(sql):
    """
    True if the sql statement is concurrent and cannot be ran in a transaction
    """
    sql = sql.strip().lower() if sql else ""
    return sql.startswith("create") and "concurrently" in sql


def _json_encoder():
    """Retrieve the JSON encoder used for encoding context"""
    class_path = getattr(
        settings, "PGHISTORY_JSON_ENCODER", "django.core.serializers.json.DjangoJSONEncoder"
    )
    return import_string(class_path)


def _inject_history_context(execute, sql, params, many, context):
    cursor = context["cursor"]

    # A named cursor automatically prepends
    # "NO SCROLL CURSOR WITHOUT HOLD FOR" to the query, which
    # causes invalid SQL to be generated. There is no way
    # to override this behavior in psycopg2, so context tracking
    # cannot happen for named cursors. Django only names cursors
    # for iterators and other statements that read the database,
    # so it seems to be safe to ignore named cursors.
    #
    # Concurrent index creation is also incompatible with local variable
    # setting. Ignore these cases for now.
    if not cursor.name and not _is_concurrent_statement(sql):
        encoder = _json_encoder()
        # Metadata is stored as a serialized JSON string with escaped
        # single quotes
        metadata_str = json.dumps(_tracker.value.metadata, cls=encoder).replace("'", "''")

        sql = (
            f"SET LOCAL pghistory.context_id='{_tracker.value.id}';"
            f"SET LOCAL pghistory.context_metadata='{metadata_str}';"
        ) + sql

    return execute(sql, params, many, context)


class context(contextlib.ContextDecorator):
    """
    A context manager that groups changes under the same context and
    adds additional metadata about the event.

    Context is added as variables at the beginning of every SQL statement.
    By default, all variables are localized to the transaction (i.e
    SET LOCAL), meaning they will only persist for the statement/transaction
    and not across the session.

    Once any code has entered ``pghistory.context``, all subsequent
    entrances of ``pghistory.context`` will be grouped under the same
    context until the top-most parent exits.

    To add context only if a parent has already entered ``pghistory.context``,
    one can call ``pghistory.context`` as a function without entering it.
    The metadata set in the function call will be part of the context if
    ``pghistory.context`` has previously been entered. Otherwise it will
    be ignored.

    Usage:

        with pghistory.context(key='value'):
            # Do things..
            # All tracked events will have the same ``pgh_context``
            # foreign key, and the context object will include
            # {'key': 'value'} in its metadata.
            # Nesting the tracker adds additional metadata to the current
            # context

        # Add metadata if a parent piece of code has already entered
        # pghistory.context
        pghistory.context(key='value')

    Args:
        metadata (dict): Metadata that should be attached to the tracking
            context

    Notes:
        Context tracking is compatible for most scenarios, but it currently
        does not work for named cursors. Django uses named cursors for
        the .iterator() operator, which has no effect on history tracking.
        However, there may be other usages of named cursors in Django where
        history context is ignored.
    """

    def __init__(self, **metadata):
        self.metadata = metadata
        self._pre_execute_hook = None

        if hasattr(_tracker, "value"):
            _tracker.value.metadata.update(**self.metadata)

    def __enter__(self):
        if not hasattr(_tracker, "value"):
            self._pre_execute_hook = connection.execute_wrapper(_inject_history_context)
            self._pre_execute_hook.__enter__()
            _tracker.value = Context(id=uuid.uuid4(), metadata=self.metadata)

        return _tracker.value

    def __exit__(self, *exc):
        if self._pre_execute_hook:
            delattr(_tracker, "value")
            self._pre_execute_hook.__exit__(*exc)



import copy
import re
import sys
import uuid

import django
from django.apps import apps
from django.db import connection
from django.db import connections
from django.db import models
from django.db.models.fields.related import RelatedField
from django.db.models.sql import Query
from django.db.models.sql.compiler import SQLCompiler

import pghistory.constants
import pghistory.trigger

# Django>=3.1 changes the location of JSONField
if (django.VERSION[0] >= 3 and django.VERSION[1] >= 1) or django.VERSION[0] >= 4:
    from django.db.models import JSONField
else:
    from django.contrib.postgres.fields import JSONField


# Create a consistent load path for JSONField regardless of django
# version. This is just to prevent migration issues for people
# on different django versions
class PGHistoryJSONField(JSONField):
    pass


class Context(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)
    metadata = PGHistoryJSONField(default=dict)

    @classmethod
    def install_pgh_attach_context_func(cls):
        """
        Installs a custom store procedure for upserting context
        for historical events. The upsert is aware of when tracking is
        enabled in the app (i.e. using pghistory.context())

        This stored procedure is automatically installed in pghistory.apps
        after migration
        """
        with connection.cursor() as cursor:
            cursor.execute(
                f"""
                CREATE OR REPLACE FUNCTION _pgh_attach_context()
                RETURNS {cls._meta.db_table}.id%TYPE AS $$
                    DECLARE
                        _pgh_context_id UUID;
                        _pgh_context_metadata JSONB;
                    BEGIN
                        BEGIN
                            SELECT INTO _pgh_context_id
                                CURRENT_SETTING('pghistory.context_id');
                            SELECT INTO _pgh_context_metadata
                                CURRENT_SETTING('pghistory.context_metadata');
                            EXCEPTION WHEN OTHERS THEN
                        END;
                        IF _pgh_context_id IS NOT NULL AND _pgh_context_metadata IS NOT NULL THEN
                            INSERT INTO {cls._meta.db_table} (id, metadata, created_at, updated_at)
                                VALUES (_pgh_context_id, _pgh_context_metadata, NOW(), NOW())
                                ON CONFLICT (id) DO UPDATE
                                    SET metadata = EXCLUDED.metadata,
                                        updated_at = EXCLUDED.updated_at;
                            RETURN _pgh_context_id;
                        ELSE
                            RETURN NULL;
                        END IF;
                    END;
                $$ LANGUAGE plpgsql;
                """
            )


def _generate_event_model_name(base_model, tracked_model, fields):
    """Generates a default history model name"""
    name = tracked_model._meta.object_name
    if fields:
        name += "_" + "_".join(fields)

    name += f"_{base_model._meta.object_name.lower()}"
    return _pascalcase(name)


def _generate_history_field(tracked_model, field):
    """
    When generating a history model from a tracked model, ensure the fields
    are set up properly so that related names and other information
    from the tracked model do not cause errors.
    """
    field = copy.deepcopy(tracked_model._meta.get_field(field))

    if isinstance(field, models.AutoField):
        field = models.IntegerField()
    elif isinstance(field, models.OneToOneField):
        field.__class__ = models.ForeignKey

    if isinstance(field, RelatedField):
        field.db_constraint = False
        field.remote_field.on_delete = models.DO_NOTHING
        field.remote_field.related_name = "+"
        field.remote_field.related_query_name = "+"
    else:
        field.db_index = False

    field.primary_key = False
    field._unique = False

    return field


def _generate_related_name(base_model, fields):
    """
    Generates a related name to the tracking model based on the base
    model and traked fields
    """
    related_name = base_model._meta.object_name.lower()
    return "_".join(fields) + f"_{related_name}" if fields else related_name


def _validate_event_model_path(*, app_label, name, abstract):
    if app_label not in apps.app_configs:
        raise ValueError(f'App label "{app_label}" is invalid')

    app = apps.app_configs[app_label]
    models_module = app.module.__name__ + ".models"
    if not abstract and hasattr(sys.modules[models_module], name):
        raise ValueError(
            f"App {app_label} already has {name} model. You must"
            " explicitly declare an unused model name for the pghistory model."
        )
    elif models_module.startswith("django."):
        raise ValueError(
            "A history model cannot be generated under third party app"
            f' "{app_label}". You must explicitly pass an app label'
            " when configuring tracking."
        )


def create_event_model(
    base_class,
    tracked_model,
    fields=None,
    exclude=None,
    obj_fk=pghistory.constants.unset,
    context_fk=pghistory.constants.unset,
    abstract=True,
    related_name=None,
    name=None,
    app_label=None,
    attrs=None,
    meta=None,
):
    """
    The primary factory function for dynamically creating a history
    model.

    Args:
        base_class (models.Model): The base class from which the created
            model will inherit
        tracked_model (models.Model): The model being tracked
        fields (List[str], default=None): The fields to track. If None,
            all fields on tracked_model are tracked.
        exclude (List[str], default=None): If no fields are provided, exclude
            these fields from tracking.
        obj_fk (models.ForeignKey, default=unset): For overriding the
            foreign key that references the tracked object. If unset, defaults
            to a non-nullable foreign key that cascades. The related name
            defaults to ``related_name`` if all fields are tracked. Otherwise
            defaults to a combintion of ``related_name`` and ``fields``. Use
            ``None`` to create a tracking model without a foreign key to
            the tracked model.
        context_fk (models.ForeignKey, default=unset): The foreign key to
            tracked context, if any. Use ``None`` to avoid attaching historical
            context altogether.
        abstract (bool, default=True): True if the created model should be
            abstract
        related_name (str): The primary way to identify the relation of
            the created model and the tracked model
        name (str, default=None): The name of the created model. If None,
            defaults to a combination of the ``related_name``,
            ``tracked_model``, and ``fields``.
        app_label (str, default=None): The app_label for the created model.
            Defaults to the app_label of ``tracked_model``. Note, when tracking
            a Django model (User) or a model of a third-party app, one must
            manually specify the app_label of an internal app to use for
            the tracking model.
        attrs (dict, default=None): Additional attributes to add to the created
            model.
        meta (dict, default=None): Additional options to add to the model
            Meta
    """
    related_name = related_name or _generate_related_name(base_class, fields)
    name = name or _generate_event_model_name(base_class, tracked_model, fields)
    app_label = app_label or tracked_model._meta.app_label
    _validate_event_model_path(app_label=app_label, name=name, abstract=abstract)
    app = apps.app_configs[app_label]
    models_module = app.module.__name__ + ".models"

    attrs = attrs or {}
    meta = meta or {}
    context_fk = (
        models.ForeignKey(
            Context,
            null=True,
            on_delete=models.DO_NOTHING,
            related_name="+",
            db_constraint=False,
        )
        if context_fk is pghistory.constants.unset
        else context_fk
    )

    obj_fk = (
        models.ForeignKey(
            tracked_model,
            null=False,
            on_delete=models.DO_NOTHING,
            db_constraint=False,
            related_name=related_name,
        )
        if obj_fk is pghistory.constants.unset
        else obj_fk
    )
    exclude = exclude or []
    fields = fields or [f.name for f in tracked_model._meta.fields if f.name not in exclude]

    class_attrs = {
        "__module__": models_module,
        "Meta": type("Meta", (), {"abstract": abstract, "app_label": app_label, **meta}),
        "pgh_tracked_model": tracked_model,
        **{field: _generate_history_field(tracked_model, field) for field in fields},
        **attrs,
    }

    if context_fk:
        class_attrs["pgh_context"] = context_fk

    if obj_fk:
        class_attrs["pgh_obj"] = obj_fk

    event_model = type(name, (base_class,), class_attrs)
    if not abstract:
        setattr(sys.modules[models_module], name, event_model)

    return event_model


def _pascalcase(string):
    """Convert string into pascal case."""

    string = re.sub(r"^[\-_\.]", "", str(string))
    if not string:  # pragma: no branch
        return string

    return string[0].upper() + re.sub(
        r"[\-_\.\s]([a-z])",
        lambda matched: matched.group(1).upper(),
        string[1:],
    )


class Event(models.Model):
    """
    An abstract model for base elements of a event
    """

    pgh_id = models.AutoField(primary_key=True)
    pgh_created_at = models.DateTimeField(auto_now_add=True)
    pgh_label = models.TextField(help_text="The event label.")
    pgh_events = None
    pgh_tracked_model = None

    class Meta:
        abstract = True

    @classmethod
    def pghistory_setup(cls):
        """
        Called when the class is prepared (see apps.py)
        to finalize setup of the model and register triggers
        """
        if not cls._meta.abstract or not cls._meta.managed:  # pragma: no branch
            for event in cls.pgh_events or []:
                event.setup(cls)

    @classmethod
    def factory(
        cls,
        model,
        *events,
        fields=None,
        exclude=None,
        obj_fk=pghistory.constants.unset,
        context_fk=pghistory.constants.unset,
        abstract=True,
        related_name=None,
        name=None,
        app_label=None,
    ):
        return create_event_model(
            cls,
            model,
            fields=fields,
            exclude=exclude,
            obj_fk=obj_fk,
            context_fk=context_fk,
            abstract=abstract,
            related_name=related_name,
            name=name,
            app_label=app_label,
            attrs={"pgh_events": events},
        )


class AggregateEventQueryCompiler(SQLCompiler):
    def _get_empty_aggregate_event_select(self):
        """
        When targetting a model that has no event tables, there are
        no valid tables from which a CTE can be generated.

        This method generates a CTE that returns an empty table in the
        schema of the AggregateEvent table. Note that it's impossible to
        create an empty CTE, so we select NULL VALUES and LIMIT to 0.
        """
        col_name_clause = ", ".join([field.column for field in self.query.model._meta.fields])
        col_select_clause = ",\n".join(
            [
                f"_pgh_obj_event.{field.column}::"
                f"{field.rel_db_type(self.connection)} AS {field.attname}"
                for field in self.query.model._meta.fields
            ]
        )
        values_list = ["(NULL)" for field in self.query.model._meta.fields]
        return f"""
            SELECT
              {col_select_clause}
            FROM (
              VALUES ({', '.join(values_list)}) LIMIT 0
            ) AS _pgh_obj_event({col_name_clause})
            WHERE pgh_table IS NOT NULL
        """

    def _class_for_target(self, obj):
        if isinstance(obj, models.QuerySet):
            return obj.model
        elif isinstance(obj, list):
            return obj[0].__class__
        return obj.__class__

    def _get_aggregate_event_select(self, obj, event_model):
        cls = self._class_for_target(obj)
        related_fields = [
            field.column
            for field in event_model._meta.fields
            if getattr(field, "related_model", None) == cls
        ]
        if not related_fields:
            raise ValueError(f"Event model {event_model} does not reference {cls}")

        event_table = event_model._meta.db_table
        if isinstance(obj, models.QuerySet) or isinstance(obj, list):
            opt = "IN"
            pks = "','".join(f"{o.pk}" for o in obj)
            pks = f"('{pks}')"
        else:
            opt = "="
            pks = f"'{obj.pk}'"
        where_filter = " OR ".join(f"_event.{col} {opt} {pks}" for col in related_fields)

        context_join_clause = ""
        final_context_columns_clause = "".join(
            [
                f"_pgh_obj_event.{field.column},\n"
                for field in self.query.model._meta.fields
                if not field.attname.startswith("pgh_")
            ]
        )
        if hasattr(event_model, "pgh_context_id"):
            context_column_clause = "pgh_context_id"

            # If the aggregate event model has any non-pgh fields,
            # pull these directly from the context metadata
            annotated_context_columns_clause = "".join(
                [
                    f"(_pgh_context.metadata->>'{field.name}')::"
                    f"{field.rel_db_type(self.connection)} AS {field.column},\n"
                    for field in self.query.model._meta.fields
                    if not field.attname.startswith("pgh_")
                ]
            )
            if annotated_context_columns_clause:
                context_join_clause = f"""
                    LEFT OUTER JOIN {Context._meta.db_table} _pgh_context
                        ON _pgh_context.id = _event.pgh_context_id
                """
        else:
            context_column_clause = "NULL::uuid AS pgh_context_id"

            # If the aggregate event model has any non-pgh fields,
            # make them null since there is no context on this event
            annotated_context_columns_clause = "".join(
                [
                    f"NULL::{field.rel_db_type(self.connection)}" f" AS {field.attname},\n"
                    for field in self.query.model._meta.fields
                    if not field.attname.startswith("pgh_")
                ]
            )

        prev_data_clause = """
            LAG(row_to_json(_event))
              OVER (
                PARTITION BY _event.pgh_obj_id, _event.pgh_label
                ORDER BY _event.pgh_id
              ) AS _prev_data
        """
        if not hasattr(event_model, "pgh_obj_id"):
            prev_data_clause = "NULL::jsonb AS _prev_data"

        return f"""
            SELECT
              _pgh_obj_event.pgh_id,
              _pgh_obj_event.pgh_created_at,
              _pgh_obj_event.pgh_label,
              {final_context_columns_clause}
              '{event_table}' AS pgh_table,
              (
                  SELECT jsonb_object_agg(filtered.key, filtered.value)
                  FROM
                    (
                        SELECT key, value
                        FROM jsonb_each(_pgh_obj_event._curr_data::jsonb)
                    ) filtered
                  WHERE filtered.key NOT LIKE 'pgh_%%'
              ) AS pgh_data,
              (
                SELECT jsonb_object_agg(curr.key, array[prev.value, curr.value])
                FROM
                  (
                    SELECT key, value
                    FROM jsonb_each(_pgh_obj_event._curr_data::jsonb)
                  ) curr
                  LEFT OUTER JOIN
                  (
                    SELECT key, value
                    FROM jsonb_each(_pgh_obj_event._prev_data::jsonb)
                  ) prev
                  ON curr.key = prev.key
                WHERE curr.key NOT LIKE 'pgh_%%'
                  AND curr.value != prev.value
                  AND prev IS NOT NULL
              ) AS pgh_diff,
              _pgh_obj_event.pgh_context_id
            FROM (
              SELECT
                pgh_id,
                pgh_created_at,
                pgh_label,
                row_to_json(_event) AS _curr_data,
                {annotated_context_columns_clause}
                {prev_data_clause},
                {context_column_clause}
              FROM {event_table} _event
              {context_join_clause}
              WHERE {where_filter}
            ) _pgh_obj_event
        """

    def get_aggregate_event_cte(self):
        """
        Returns the CTE clause for the aggregate event query
        """
        obj = self.query.target
        if not obj:
            raise ValueError("Must use .target() to target an object for event aggregation")

        event_models = self.query.across
        cls = self._class_for_target(obj)
        if not event_models:
            event_models = [
                model
                for model in apps.get_models()
                if issubclass(model, Event)
                and not issubclass(model, BaseAggregateEvent)
                and any(
                    getattr(field, "related_model", None) == cls for field in model._meta.fields
                )
            ]

        agg_event_table = self.query.model._meta.db_table
        inner_cte = "UNION ALL ".join(
            [self._get_aggregate_event_select(obj, event_model) for event_model in event_models]
        )
        if not inner_cte:
            inner_cte = self._get_empty_aggregate_event_select()

        return f"WITH {agg_event_table} AS (\n" + inner_cte + "\n)\n"

    def as_sql(self, *args, **kwargs):
        base_sql, base_params = super().as_sql(*args, **kwargs)

        # Create the CTE that will be queried and insert it into the
        # main query
        cte = self.get_aggregate_event_cte()

        return cte + base_sql, base_params


class AggregateEventQuery(Query):
    """A query over an aggregate event CTE"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.target = None
        self.across = []

    def get_compiler(self, using=None, connection=None):  # pragma: no cover
        """
        Overrides the Query method get_compiler in order to return
        an AggregateEventCompiler.

        Copies the body of Django's get_compiler and overrides the return,
        so we ignore covering this method.
        """
        # Copy the body of this method from Django except the final
        # return statement.
        if using is None and connection is None:
            raise ValueError("Need either using or connection")

        if using:
            connection = connections[using]

        # Check that the compiler will be able to execute the query
        for _, aggregate in self.annotation_select.items():
            connection.ops.check_expression_support(aggregate)

        # Instantiate the custom compiler.
        return AggregateEventQueryCompiler(self, connection, using)

    def __chain(self, _name, klass=None, *args, **kwargs):
        clone = getattr(super(), _name)(self.__class__, *args, **kwargs)
        clone.target = self.target
        clone.across = self.across
        return clone

    if django.VERSION < (2, 0):  # pragma: no cover

        def clone(self, klass=None, *args, **kwargs):
            return self.__chain("clone", klass, *args, **kwargs)

    else:

        def chain(self, klass=None):
            return self.__chain("chain", klass)


class AggregateEventQuerySet(models.QuerySet):
    """QuerySet with support for Common Table Expressions"""

    def __init__(self, model=None, query=None, using=None, hints=None):
        # Only create an instance of a Query if this is the first invocation in
        # a query chain.
        if query is None:
            query = AggregateEventQuery(model)
        super().__init__(model, query, using, hints)

    def across(self, *event_models):
        """Aggregates events across the provided event models"""
        qs = self._clone()
        qs.query.across = event_models
        return qs

    def target(self, obj):
        """Target an object to aggregate events against"""
        qs = self._clone()
        qs.query.target = obj
        return qs


class BaseAggregateEvent(Event):
    """
    A proxy model for aggregating events together across tables and
    rendering diffs
    """

    pgh_table = models.CharField(
        max_length=64, help_text="The table under which the event is stored"
    )
    pgh_data = PGHistoryJSONField(help_text="The raw data of the event row")
    pgh_diff = PGHistoryJSONField(
        help_text="The diff between the previous event and the current event"
    )
    pgh_context = models.ForeignKey(
        Context,
        null=True,
        help_text="The context, if any, associated with the event",
        on_delete=models.DO_NOTHING,
    )

    objects = AggregateEventQuerySet.as_manager()

    class Meta:
        abstract = True


class AggregateEvent(BaseAggregateEvent):
    """
    A proxy model for aggregating events together across tables and
    rendering diffs
    """

    class Meta:
        managed = False



import django

from pghistory.core import AfterInsert
from pghistory.core import AfterInsertOrUpdate
from pghistory.core import AfterUpdate
from pghistory.core import BeforeDelete
from pghistory.core import BeforeUpdate
from pghistory.core import create_event
from pghistory.core import DatabaseEvent
from pghistory.core import Event
from pghistory.core import get_event_model
from pghistory.core import Snapshot
from pghistory.core import track
from pghistory.tracking import context
from pghistory.version import __version__


__all__ = [
    "AfterInsert",
    "AfterInsertOrUpdate",
    "AfterUpdate",
    "BeforeDelete",
    "BeforeUpdate",
    "context",
    "create_event",
    "DatabaseEvent",
    "get_event_model",
    "Snapshot",
    "Event",
    "track",
    "__version__",
]

if django.VERSION < (3, 2):
    default_app_config = "pghistory.apps.PGHistoryConfig"

del django



from django.conf import settings
from django.core.handlers.wsgi import WSGIRequest as DjangoWSGIRequest

import pghistory


class WSGIRequest(DjangoWSGIRequest):
    """
    Although Django's auth middleware sets the user in middleware,
    apps like django-rest-framework set the user in the view layer.
    This creates issues for pghistory tracking since the context needs
    to be set before DB operations happen.

    This special WSGIRequest updates pghistory context when
    the request.user attribute is updated.
    """

    def __setattr__(self, attr, value):
        if attr == "user":
            pghistory.context(user=value.pk if value and hasattr(value, "pk") else None)

        return super().__setattr__(attr, value)


def HistoryMiddleware(get_response):
    """
    Annotates the user/url in the pghistory context.
    """

    def middleware(request):
        middleware_methods = getattr(
            settings, "PGHISTORY_MIDDLEWARE_METHODS", ("GET", "POST", "PUT", "PATCH", "DELETE")
        )

        if request.method in middleware_methods:
            user = (
                request.user.pk
                if hasattr(request, "user") and hasattr(request.user, "pk")
                else None
            )
            with pghistory.context(user=user, url=request.path):
                if isinstance(request, DjangoWSGIRequest):  # pragma: no branch
                    request.__class__ = WSGIRequest

                return get_response(request)
        else:
            return get_response(request)

    return middleware



from django.db import models
import pgtrigger


def _get_pgh_obj_pk_col(history_model):
    """
    Returns the column name of the PK field tracked by the history model
    """
    return history_model._meta.get_field("pgh_obj").related_model._meta.pk.column


class Event(pgtrigger.Trigger):
    """
    Events a model with a label when a condition happens
    """

    label = None
    snapshot = "NEW"
    event_model = None
    when = pgtrigger.After

    def __init__(
        self,
        *,
        name=None,
        operation=None,
        condition=None,
        when=None,
        label=None,
        snapshot=None,
        event_model=None,
    ):
        self.label = label or self.label
        if not self.label:  # pragma: no cover
            raise ValueError('Must provide "label"')

        self.event_model = event_model or self.event_model
        if not self.event_model:  # pragma: no cover
            raise ValueError('Must provide "event_model"')

        self.snapshot = snapshot or self.snapshot
        if not self.snapshot:  # pragma: no cover
            raise ValueError('Must provide "snapshot"')

        super().__init__(name=name, operation=operation, condition=condition, when=when)

    def get_func(self, model):
        fields = {
            f.column: f'{self.snapshot}."{f.column}"'
            for f in self.event_model._meta.fields
            if not isinstance(f, models.AutoField)
            and hasattr(self.event_model.pgh_tracked_model, f.name)
        }
        fields["pgh_created_at"] = "NOW()"
        fields["pgh_label"] = f"'{self.label}'"

        if hasattr(self.event_model, "pgh_obj"):
            fields["pgh_obj_id"] = f'{self.snapshot}."{_get_pgh_obj_pk_col(self.event_model)}"'

        if hasattr(self.event_model, "pgh_context"):
            fields["pgh_context_id"] = "_pgh_attach_context()"

        cols = ", ".join(f'"{col}"' for col in fields)
        vals = ", ".join(val for val in fields.values())
        sql = f"""
            INSERT INTO "{self.event_model._meta.db_table}"
                ({cols}) VALUES ({vals});
            RETURN NULL;
        """
        return " ".join(line.strip() for line in sql.split("\n") if line.strip()).strip()



# Generated by Django 3.0.7 on 2020-06-23 14:53

import django.contrib.postgres.fields.jsonb
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("pghistory", "0001_initial")]

    operations = [
        migrations.CreateModel(
            name="AggregateEvent",
            fields=[
                (
                    "pgh_id",
                    models.AutoField(primary_key=True, serialize=False),
                ),
                ("pgh_created_at", models.DateTimeField(auto_now_add=True)),
                ("pgh_label", models.TextField(help_text="The event label.")),
                (
                    "pgh_table",
                    models.CharField(
                        help_text="The table under which the event is stored",
                        max_length=64,
                    ),
                ),
                (
                    "pgh_data",
                    django.contrib.postgres.fields.jsonb.JSONField(
                        help_text="The raw data of the event row"
                    ),
                ),
                (
                    "pgh_diff",
                    django.contrib.postgres.fields.jsonb.JSONField(
                        help_text="The diff between the previous event and the current event"
                    ),
                ),
            ],
            options={"managed": False},
        )
    ]



import django.apps
from django.db.models.signals import class_prepared
from django.db.models.signals import post_migrate


def install_pgh_attach_context_func(**kwargs):
    """
    Installs a special stored procedure used by pghistory that makes
    it easier to manually and automatically construct context for
    history
    """
    Context = django.apps.apps.get_model("pghistory", "Context")
    Context.install_pgh_attach_context_func()


def pgh_setup(sender, **kwargs):
    if hasattr(sender, "pghistory_setup"):
        sender.pghistory_setup()


class PGHistoryConfig(django.apps.AppConfig):
    name = "pghistory"

    def __init__(self, *args, **kwargs):
        """
        Install pgtriggers for Change detection models whenever
        the change detection model class is ready. We have to do this
        in __init__ instead of ready() here since the class_prepared
        signal is emitted before models are ready
        """
        class_prepared.connect(pgh_setup)
        super().__init__(*args, **kwargs)

    def ready(self):
        post_migrate.connect(install_pgh_attach_context_func, sender=self)



# Generated by Django 3.0.6 on 2020-05-28 18:37
import uuid

import django.contrib.postgres.fields.jsonb
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="Context",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("updated_at", models.DateTimeField(auto_now=True)),
                (
                    "metadata",
                    django.contrib.postgres.fields.jsonb.JSONField(default=dict),
                ),
            ],
        )
    ]



# Generated by Django 3.1.2 on 2020-10-23 21:36

from django.db import migrations

import pghistory.models


class Migration(migrations.Migration):

    dependencies = [
        ("pghistory", "0002_aggregateevent"),
    ]

    operations = [
        migrations.AlterField(
            model_name="context",
            name="metadata",
            field=pghistory.models.PGHistoryJSONField(default=dict),
        ),
    ]



"""Core functionality and interface of pghistory"""
import re

from django.db import connection
from django.db import models
from django.db.models import sql
from django.db.models.sql import compiler
import pgtrigger
from psycopg2.extensions import AsIs

import pghistory.constants


_registered_events = {}


def _get_name_from_label(label):
    """Given a history event label, generate a trigger name"""
    if label:
        return re.sub("[^0-9a-zA-Z]+", "_", label)
    else:  # pragma: no cover
        return None


class Event:
    """For storing an event when a condition happens on a model

    Events that inherit this base class are assumed to be
    manually created by the user. Only a "label" for the event is
    required.

    Events that are automatically created in a Postgres trigger should
    inherit `DatabaseEvent`
    """

    label = None

    def __init__(self, label=None):
        self.label = label or self.label
        if not self.label:
            raise ValueError(f'{self.__class__.__name__} must have "label" attribute')

    def setup(self, event_model):
        """Set up the event for the particular event model"""
        pass


class DatabaseEvent(Event):
    """For tracking an event automatically based on database changes."""

    when = None
    condition = None
    operation = None
    snapshot = None

    def __init__(
        self,
        label=None,
        *,
        when=None,
        condition=None,
        operation=None,
        snapshot=None,
    ):
        super().__init__(label=label)

        self.when = when or self.when
        self.condition = condition or self.condition
        self.operation = operation or self.operation
        self.snapshot = snapshot or self.snapshot

    def setup(self, event_model):
        pgtrigger.register(
            pghistory.trigger.Event(
                event_model=event_model,
                label=self.label,
                name=_get_name_from_label(self.label),
                snapshot=self.snapshot,
                when=self.when,
                operation=self.operation,
                condition=self.condition,
            )
        )(event_model.pgh_tracked_model)


class Snapshot(DatabaseEvent):
    """
    A special database event that tracks changes to fields.
    A snapshot event always fires for an insert and also fires
    for updates when any fields change.

    NOTE: Two triggers must be created since Insert triggers do
    not allow comparison against the OLD values. We could also
    place this in one trigger and do the condition in the plpgsql code.
    """

    def __init__(self, label=None):
        return super().__init__(label=label)

    def setup(self, event_model):

        insert_trigger = pghistory.trigger.Event(
            event_model=event_model,
            label=self.label,
            name=_get_name_from_label(f"{self.label}_insert"),
            snapshot="NEW",
            when=pgtrigger.After,
            operation=pgtrigger.Insert,
        )

        condition = pgtrigger.Q()
        for field in event_model._meta.fields:
            if hasattr(event_model.pgh_tracked_model, field.name):
                condition |= pgtrigger.Q(
                    **{f"old__{field.name}__df": pgtrigger.F(f"new__{field.name}")}
                )

        update_trigger = pghistory.trigger.Event(
            event_model=event_model,
            label=self.label,
            name=_get_name_from_label(f"{self.label}_update"),
            snapshot="NEW",
            when=pgtrigger.After,
            operation=pgtrigger.Update,
            condition=condition,
        )

        pgtrigger.register(insert_trigger, update_trigger)(event_model.pgh_tracked_model)


class PreconfiguredDatabaseEvent(DatabaseEvent):
    """
    A base database event that only takes a condition. Subclasses
    preconfigure the other parameters
    """

    def __init__(self, label=None, *, condition=None):
        return super().__init__(label=label, condition=condition)


class AfterInsertOrUpdate(PreconfiguredDatabaseEvent):
    """
    A database event that happens after insert/update
    """

    operation = pgtrigger.Insert | pgtrigger.Update
    snapshot = "NEW"


class AfterInsert(PreconfiguredDatabaseEvent):
    """For events that happen after a database insert"""

    operation = pgtrigger.Insert
    snapshot = "NEW"


class BeforeUpdate(PreconfiguredDatabaseEvent):
    """
    For events that happen before a database update. The OLD values of the row
    will be snapshot to the event model
    """

    operation = pgtrigger.Update
    snapshot = "OLD"


class AfterUpdate(PreconfiguredDatabaseEvent):
    """
    For events that happen after a database update. The NEW values of the row
    will be snapshot to the event model
    """

    operation = pgtrigger.Update
    snapshot = "NEW"


class BeforeDelete(PreconfiguredDatabaseEvent):
    """
    For events that happen before a database deletion.
    """

    operation = pgtrigger.Delete
    snapshot = "OLD"


def get_event_model(
    model,
    *events,
    fields=None,
    exclude=None,
    obj_fk=pghistory.constants.unset,
    context_fk=pghistory.constants.unset,
    related_name=None,
    name=None,
    app_label=None,
    abstract=True,
):
    """
    Obtain a base event model.

    Instead of using `pghistory.track`, which dynamically generates an event
    model, one can instead construct a event model themselves, which
    will also set up event tracking for the original model.

    Usage:

        class MyEventModel(get_event_model(
            TrackedModel,
            pghistory.AfterInsert('model_create'),
        )):
            # Add custom indices or change default field declarations...

    Args:
        model (models.Model): The model that is being tracked.
        *events (List[`Event`]): See "events" help from
            `pghistory.track`.
        fields (List[str], default=None): See "fields" help from
            `pghistory.track`.
        exclude (List[str], default=None): See "exclude" help from
            `pghistory.track`.
        obj_fk (models.ForeignKey): See "obj_fk" help from
            `pghistory.track`.
        related_name (str, default=None): See "related_name" help from
            `pghistory.track`.
        name (str, default=None): See "model_name" help from
            `pghistory.track`.
        app_label (str, default=None): See "app_label" help
            from `pghistory.track`.
        abstract (bool, default=True): ``True`` if the generated model should
            be an abstract model.
    """
    # Avoid importing models in the core module since these functions
    # are imported at the top level of the package
    import pghistory.models

    event_model = pghistory.models.Event.factory(
        model,
        *events,
        fields=fields,
        exclude=exclude,
        obj_fk=obj_fk,
        context_fk=context_fk,
        related_name=related_name,
        name=name,
        app_label=app_label,
        abstract=abstract,
    )

    return event_model


def track(
    *events,
    fields=None,
    exclude=None,
    obj_fk=pghistory.constants.unset,
    context_fk=pghistory.constants.unset,
    related_name=None,
    model_name=None,
    app_label=None,
):
    """
    A decorator for tracking events for a mdoel.

    When using this decorator, an event model is dynamically generated
    that snapshots the entire model or supplied fields of the model
    based on the ``events`` supplied. The snapshot is accompanied with
    the label that identifies the event.

    Args:
        *events (List[`Event`]): The events to track. When using any event that
            inherits `pghistory.DatabaseEvent`, such as
            `pghistory.AfterInsert`, a Postgres trigger will be installed that
            automatically tracks the event with a generated event model. Events
            that do not inherit `pghistory.DatabaseEvent` are assumed to be
            manually tracked by the user.
        fields (List[str], default=None): The list of fields to snapshot
            when the event takes place. When no fields are provided, the entire
            model is snapshot when the event happens. Note that snapshotting
            of the OLD or NEW row is configured by the ``snapshot``
            attribute of the `DatabaseEvent` object. Manual events must specify
            these fields during manual creation.
        exclude (List[str], default=None): Instead of providing a list
            of fields to snapshot, a user can instead provide a list of fields
            to not snapshot.
        obj_fk (models.ForeignKey, default=unset): The foreign key field that
            references the eventged object. Defaults to a non-nullable foreign
            key that cascade deletes.  Use ``None`` to create a event model
            with no reference to the tracked object. See ``related_name``
            attribute for how the related_name is determined.
        context_fk (models.ForeignKey, default=unset): The foreign key to
            tracked context, if any. Use ``None`` to avoid attaching historical
            context altogether.
        related_name (str, default=None): The related name of the event
            model. If not provided, defaults to "event" if one is tracking
            changes to the entire model, otherwise defaults to a name based on
            the combination of fields.
        model_name (str, default=None): Use a custom model name
            when the event model is generated. Otherwise a default
            name based on the tracked model and fields will be created.
        app_label (str, default=None): The app_label for the generated
            event model. Defaults to the app_label of the tracked model. Note,
            when tracking a Django model (User) or a model of a third-party
            app, one must manually specify the app_label of an internal app to
            use so that migrations work properly.
    """

    def _model_wrapper(model_class):
        event_model = get_event_model(
            model_class,
            *events,
            fields=fields,
            exclude=exclude,
            obj_fk=obj_fk,
            context_fk=context_fk,
            name=model_name,
            related_name=related_name,
            app_label=app_label,
            abstract=False,
        )
        for event in events:
            _registered_events[(model_class, event.label)] = event_model

        return model_class

    return _model_wrapper


class _InsertEventCompiler(compiler.SQLInsertCompiler):
    def as_sql(self, *args, **kwargs):
        ret = super().as_sql(*args, **kwargs)
        assert len(ret) == 1
        params = [
            param if field.name != "pgh_context" else AsIs("_pgh_attach_context()")
            for field, param in zip(self.query.fields, ret[0][1])
        ]
        return [(ret[0][0], params)]


def create_event(obj, *, label, using="default"):
    """Manually create a event for an object.

    Events are automatically linked with any context being tracked
    via `pghistory.context`.

    Args:
        obj (models.Model): An instance of a model.
        label (str): The event label.

    Raises:
        ValueError: If the event label has not been registered for the model.

    Returns:
        models.Model: The created event model.
    """
    # Verify that the provided event is registered to the object model
    if (obj.__class__, label) not in _registered_events:
        raise ValueError(
            f'"{label}" is not a registered event for model' f" {obj._meta.object_name}."
        )

    event_model = _registered_events[(obj.__class__, label)]
    event_model_kwargs = {
        "pgh_label": label,
        **{
            field.attname: getattr(obj, field.attname)
            for field in event_model._meta.fields
            if not field.name.startswith("pgh_")
        },
    }
    if hasattr(event_model, "pgh_obj"):
        event_model_kwargs["pgh_obj"] = obj

    event_obj = event_model(**event_model_kwargs)

    # The event model is inserted manually with a custom SQL compiler
    # that attaches the context using the _pgh_attach_context
    # stored procedure. Django does not allow one to use F()
    # objects to reference stored procedures, so we have to
    # inject it with a custom SQL compiler here.
    query = sql.InsertQuery(event_model)
    query.insert_values(
        [field for field in event_model._meta.fields if not isinstance(field, models.AutoField)],
        [event_obj],
    )

    vals = _InsertEventCompiler(query, connection, using="default").execute_sql(
        event_model._meta.fields
    )

    # Django <= 2.2 does not support returning fields from a bulk create,
    # which requires us to fetch fields again to populate the context
    # NOTE (@wesleykendall): We will eventually test multiple Django versions
    if isinstance(vals, int):  # pragma: no cover
        return event_model.objects.get(pgh_id=vals)
    else:
        # Django >= 3.1 returns the values as a list of one element
        if isinstance(vals, list) and len(vals) == 1:  # pragma: no branch
            vals = vals[0]

        for field, val in zip(event_model._meta.fields, vals):
            setattr(event_obj, field.attname, val)

        return event_obj



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pghistory', 'pghistory.migrations']

package_data = \
{'': ['*']}

install_requires = \
['django-pgtrigger>=3.2.0', 'django>=2']

extras_require = \
{':python_version >= "3.7" and python_version < "3.8"': ['importlib_metadata>=4']}

setup_kwargs = {
    'name': 'django-pghistory',
    'version': '2.2.2',
    'description': 'History tracking for Django and Postgres',
    'long_description': 'django-pghistory\n################\n\n``django-pghistory`` provides automated and customizable history\ntracking for Django models using\n`Postgres triggers <https://www.postgresql.org/docs/12/sql-createtrigger.html>`__.\nUsers can configure a number of event trackers to snapshot every model\nchange or to fire specific events when certain changes occur in the database.\n\nIn contrast with other Django auditing and history tracking apps\n(seen `here <https://djangopackages.org/grids/g/model-audit/>`__),\n``django-pghistory`` has the following advantages:\n\n1. No instrumentation of model and queryset methods in order to properly\n   track history. After configuring your model, events will be tracked\n   automatically with no other changes to code. In contrast with\n   apps like\n   `django-reversion <https://django-reversion.readthedocs.io/en/stable/>`__,\n   it is impossible for code to accidentally bypass history tracking, and users\n   do not have to use a specific model/queryset interface to ensure history\n   is correctly tracked.\n2. Bulk updates and all other modifications to the database that do not fire\n   Django signals will still be properly tracked.\n3. Historical event modeling is completely controlled by the user and kept\n   in sync with models being tracked. There are no cumbersome generic foreign\n   keys and little dependence on unstructured JSON fields for tracking changes,\n   making it easier to use the historical events in your application (and\n   in a performant manner).\n4. Changes to multiple objects in a request (or any level of granularity)\n   can be grouped together under the same context. Although history tracking\n   happens in Postgres triggers, application code can still attach metadata\n   to historical events, such as the URL of the request, leading to a more\n   clear and useful audit trail.\n\nTo get started, read the `django-pghistory docs\n<https://django-pghistory.readthedocs.io/>`__. The docs covers how to\nset up and configure automated event tracking in your application, along\nwith how to aggregate events for objects and visualize them in your\nadmin/application.\n\nDocumentation\n=============\n\n`View the django-pghistory docs here\n<https://django-pghistory.readthedocs.io/>`_.\n\nInstallation\n============\n\nInstall django-pghistory with::\n\n    pip3 install django-pghistory\n\nAfter this, add ``pghistory`` to the ``INSTALLED_APPS``\nsetting of your Django project.\n\nContributing Guide\n==================\n\nFor information on setting up django-pghistory for development and\ncontributing changes, view `CONTRIBUTING.rst <CONTRIBUTING.rst>`_.\n\nPrimary Authors\n===============\n\n- @wesleykendall (Wes Kendall, wesleykendall@protonmail.com)\n\nOther Contributors\n==================\n\n- @shivananda-sahu\n- @asucrews\n- @Azurency\n- @dracos\n- @adamchainz\n- @eeriksp\n',
    'author': 'Wes Kendall',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/Opus10/django-pghistory',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7.0,<4',
}


setup(**setup_kwargs)
